import copy
import enum
import re
from typing import Optional
from urllib.parse import urljoin

import requests
from requests import HTTPError  # re-export to allow for exception handling

from . import app_settings
from .utils import datetime_parser, logger

default_app_config = "postgrestutils.apps.PostgrestUtilsConfig"


REPR_OUTPUT_SIZE = 20

Count = enum.Enum(
    'Count',
    (
        ('NONE', None),
        ('EXACT', 'exact'),
        ('PLANNED', 'planned'),
        ('ESTIMATED', 'estimated')
    )
)

DEFAULT_SCHEMA = object()


class ObjectDoesNotExist(Exception):
    pass


class MultipleObjectsReturned(Exception):
    pass


class Session:
    """
    A PostgREST session to be used as a context manager.
    The session will default to the options specified on creation. However,
    some of these may be overridden on a per-request basis by using the `.get()`
    or `.filter()` methods.
    """
    def __init__(
        self,
        base_uri: Optional[str] = None,
        token: Optional[str] = None,
        schema: Optional[str] = None,
        parse_dt: bool = True,
        count: Count = Count.NONE
    ):
        """
        :param base_uri: base uri of the PostgREST instance to use
        :param token: JWT for the corresponding database role and PostgREST
        instance
        :param schema: the database schema to use
        :param parse_dt: whether to parse datetime strings as returned by
        PostgREST to python datetime objects
        :param count: counting strategy as explained in the README
        """
        self.session = None
        self.base_uri = base_uri or app_settings.BASE_URI
        self.token = token or app_settings.JWT
        self.schema = schema
        self.parse_dt = parse_dt
        self.count = count

    def __enter__(self):
        self.session = requests.Session()
        self._configure_session_defaults()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.session.close()

    def get(
        self,
        endpoint: str,
        parse_dt: Optional[bool] = None,
        count: Optional[Count] = None,
        schema: Optional[str] = None,
        **kwargs
    ) -> dict:
        """
        Get a single object from the specified endpoint. This will most likely
        require specifying params so that only a single object is found.
        :param endpoint: specifies which endpoint to request
        when multiple results are returned)
        :param parse_dt: whether to parse datetime strings as returned by
        PostgREST to python datetime objects
        :param count: counting strategy as explained in the README
        :param kwargs: pass kwargs directly to requests's `.get()` method
        :return: single element as dict
        :raises: `ObjectDoesNotExist`/`MultipleObjectsReturned` if no or more
        than one object was found
        """
        if schema is not None:
            self._set_schema_header(schema, kwargs)
        res = JsonResultSet(
            self,
            endpoint,
            True,
            parse_dt if parse_dt is not None else self.parse_dt,
            count if count is not None else self.count,
            **kwargs
        )
        # populate the cache
        # will raise ObjectDoesNotExist/MultipleObjectsReturned if no or
        # multiple elements are returned
        res._fetch_all()
        assert res._result_cache is not None
        return res._result_cache[0]

    def filter(
        self,
        endpoint: str,
        parse_dt: Optional[bool] = None,
        count: Optional[Count] = None,
        schema: Optional[str] = None,
        **kwargs
    ) -> 'JsonResultSet':
        """
        :param endpoint: specifies which endpoint to request
        :param parse_dt: whether to parse datetime strings as returned by
        PostgREST to python datetime objects
        :param count: counting strategy as explained in the README
        :param kwargs: pass kwargs directly to request's `.get()` method
        :return: `JsonResultSet`, a lazy python object
        """
        if schema is not None:
            self._set_schema_header(schema, kwargs)
        return JsonResultSet(
            self,
            endpoint,
            False,
            parse_dt if parse_dt is not None else self.parse_dt,
            count if count is not None else self.count,
            **kwargs
        )

    def _configure_session_defaults(self):
        self.session.headers['Accept'] = 'application/json'
        if self.token:
            self.session.headers['Authorization'] = 'Bearer {}'.format(self.token)
        if self.schema is not None:
            self.session.headers['Accept-Profile'] = self.schema

    def _set_schema_header(self, schema, kwargs: dict):
        if schema is DEFAULT_SCHEMA:
            schema = None
        kwargs.setdefault('headers', dict())['Accept-Profile'] = schema


class JsonResultSet:
    """
    A lazy python object that is returned when calling the `.filter()` method
    on a `postgrestutils.Session`.
    It behaves akin to django's `Queryset` and implements some dunder methods
    to ensure pythonic behavior.
    Check the README for more detailed information.
    """
    def __init__(self, client: Session, endpoint: str, singular: bool, parse_dt: bool, count: Count, **kwargs):
        self._len_cache = None  # type: Optional[int]
        self._result_cache = None  # type: Optional[list]

        self.client = client  # type: Session
        self.endpoint = endpoint  # type: str
        self.singular = singular  # type: bool
        self.parse_dt = parse_dt  # type: bool
        self.count = count  # type: Count
        self.request_kwargs = kwargs

    def __repr__(self):
        data = list(self[:REPR_OUTPUT_SIZE + 1])
        if len(data) > REPR_OUTPUT_SIZE:
            data[-1] = "...(remaining elements truncated)..."
        return '<{} {}>'.format(self.__class__.__name__, data)

    def __iter__(self):
        self._fetch_all()
        return iter(self._result_cache)

    def __len__(self):
        """
        NOTE: Since singular requests (using `.get()`) return a python dict
        rather than a `JsonResultSet`, `self.singular` should be ignored here.
        """
        if self.count != Count.NONE:
            self._fetch_len()
        else:
            self._fetch_all()
        return self._len_cache

    def __getitem__(self, key):
        """
        NOTE: Since singular requests (using `.get()`) return a python dict
        rather than a `JsonResultSet`, `self.singular` should be ignored here.
        """
        if not isinstance(key, (int, slice)):
            raise TypeError(
                "{self.__class__.__name__} indices must be integers or slices, not {key.__class__.__name__}".format(
                    self=self,
                    key=key
                )
            )
        if ((isinstance(key, int) and key < 0) or
           (isinstance(key, slice) and ((key.start is not None and key.start < 0) or
                                        (key.stop is not None and key.stop < 0)))):
            raise ValueError("{self.__class__.__name__} does not support negative indexing".format(self=self))
        if isinstance(key, slice) and key.step is not None:
            raise ValueError("{self.__class__.__name__} does not support stepping".format(self=self))

        # cache is not populated and unbounded slice is requested, i.e. res[:]
        if isinstance(key, slice) and all(e is None for e in (self._result_cache, key.start, key.stop)):
            self._fetch_all()

        if self._result_cache is not None:
            return self._result_cache[key]

        if isinstance(key, slice):
            start = key.start if key.start is not None else 0
            if key.stop is not None and key.stop <= start:
                return list()
            range = '{start}-{stop}'.format(start=start, stop=key.stop - 1 if key.stop is not None else '')
            return self._fetch_range(range)
        return self._fetch_range('{0}-{0}'.format(key))[0]  # single element requested, return dict

    def refresh_from_pgrest(self):
        """Lazily refresh data from PostgREST."""
        self._result_cache = None
        self._len_cache = None

    def _fetch_len(self):
        """
        Fetch the length from the PostgREST API.
        NOTE: This method should ignore `self.singular`, see `__len__()` for
        more information.
        """
        if self._len_cache is None:
            request_kwargs = copy.deepcopy(self.request_kwargs)
            request_kwargs.setdefault('headers', dict())['Prefer'] = 'count={}'.format(self.count.value)
            request_kwargs['headers']['Range-Unit'] = 'items'
            # Have to request something so just fetch the first item
            request_kwargs['headers']['Range'] = '0-0'

            resp = self.client.session.get(urljoin(self.client.base_uri, self.endpoint), **request_kwargs)

            count = int(resp.headers['Content-Range'].split('/')[-1])
            self._len_cache = count

            # If the request yields only one element anyway, might as well cache
            # it. When using singular=False, count=Count.EXACT and the result
            # is a single element this saves an API request in cases where
            # len() is called before using the result.
            if count == 1:
                self._result_cache = self._parse_response(resp)

    def _fetch_all(self):
        """
        Fetch all elements that match the specified params and populate the
        caches with results.
        """
        if self._result_cache is None:
            request_kwargs = copy.deepcopy(self.request_kwargs)

            if self.singular:
                request_kwargs.setdefault('headers', dict())['Accept'] = 'application/vnd.pgrst.object+json'

            resp = self.client.session.get(urljoin(self.client.base_uri, self.endpoint), **request_kwargs)
            self._result_cache = self._parse_response(resp)

            # fetched all elements anyway, caching their length is very cheap
            self._len_cache = len(self._result_cache)

    def _fetch_range(self, range):
        """
        Fetch a range of elements from the PostgREST API.
        NOTE: This method should ignore `self.singular`, see `__getitem__()` for
        more information.
        """
        request_kwargs = copy.deepcopy(self.request_kwargs)
        request_kwargs.setdefault('headers', dict())['Range-Unit'] = 'items'
        request_kwargs['headers']['Range'] = range

        resp = self.client.session.get(urljoin(self.client.base_uri, self.endpoint), **request_kwargs)
        return self._parse_response(resp)

    def _parse_response(self, resp):
        """
        Parse response as json and return the result if it was successful.
        Attempt to detect common error cases in order to raise meaningful error
        messages.
        :return: a list of successfully parsed objects
        :raises: `ObjectDoesNotExist`/`MultipleObjectsReturned` if `.get()` was
        used but the API couldn't return exactly one object. `HTTPError` in all
        other cases e.g. network-related issues.
        """
        try:
            resp.raise_for_status()
        except requests.HTTPError as e:
            # try getting a more detailed exception if status_code = 406
            if resp.status_code == 406:
                try:
                    self._try_parse_406(resp)
                except (ObjectDoesNotExist, MultipleObjectsReturned) as detailed:
                    raise detailed from e

            # fall back to raising a generic HTTPError exception
            raise type(e)(resp.status_code, resp.reason, resp.text, response=resp, request=e.request)

        if self.parse_dt:
            json_result = resp.json(object_hook=datetime_parser)
        else:
            json_result = resp.json()
        # always return a list even if it contains a single element only
        return [json_result] if self.singular else json_result

    def _try_parse_406(self, resp):
        """
        Try parsing a 406 `HTTPError` to raise a more detailed error message.
        :param resp: the HTTP response to parse
        :raises: `ObjectDoesNotExist`/`MultipleObjectsReturned` depending on the
        observed row count.
        """
        detail_regex = re.compile(
            r'Results contain (?P<row_count>\d+) rows, application/vnd\.pgrst\.object\+json requires 1 row'
        )
        try:
            json = resp.json()
        except ValueError:
            # Failed parsing as json, give up trying to guess the error.
            # PostgREST probably changed the error format, log the response for
            # more insights.
            logger.warning("Unparsable 406: {}".format(resp.text))
        else:
            result = re.match(detail_regex, json['details'])
            if result is not None:
                row_count = int(result.group('row_count'))
                if row_count == 0:
                    raise ObjectDoesNotExist(json)
                else:
                    raise MultipleObjectsReturned(json)
