import importlib.util
import logging
import re
from datetime import datetime, timedelta, timezone
from typing import Dict, Union

_DJANGO = importlib.util.find_spec("django") is not None
if _DJANGO:
    from ._django_utils import _try_django_parse_dt, autofetch


logger = logging.getLogger("postgrestutils")

# this regex matches postgres' JSON formatting for timestamps
JSON_TIMESTAMP_REGEX = re.compile(r"(?P<year>\d{4})-"
                                  r"(?P<month>\d{2})-"
                                  r"(?P<day>\d{2})"
                                  r"T(?P<hour>\d{2}):"
                                  r"(?P<minute>\d{2}):"
                                  r"(?P<second>\d{2})\."
                                  r"(?P<microsecond>\d{1,6})"
                                  r"((?P<offsetsign>[+-])"
                                  r"(?P<offsethours>\d{2}):"
                                  r"(?P<offsetminutes>\d{2}))?$")


def _clean_parts(parts: Dict[str, str]):
    cleaned = {}  # type: Dict[str, Union[int, str]]
    for key, value in parts.items():
        if value:
            if "offset" not in key:
                cleaned[key] = int(value)
            else:
                cleaned[key] = value
    return cleaned


def _try_python_parse_dt(value: str) -> Union[datetime, str]:
    """
    Attempt to parse value as a datetime using only python utilities.
    :param value: the string to parse
    :return: the parsed `datetime` or the original string if parsing failed
    """
    match = JSON_TIMESTAMP_REGEX.match(value)
    if match:
        parts = _clean_parts(match.groupdict())
        if parts.get('offsetsign') and parts.get('offsethours') and parts.get('offsetminutes'):
            sign = -1 if parts.pop('offsetsign', '+') == '-' else 1
            tz = timezone(
                offset=sign * timedelta(
                    hours=int(parts.pop('offsethours')),
                    minutes=int(parts.pop('offsetminutes'))
                )
            )
            parsed_dt = datetime(**parts).replace(tzinfo=tz).astimezone()
        else:
            # naive datetime so we assume local time
            local_tz = datetime.now(timezone.utc).astimezone().tzinfo
            parsed_dt = datetime(**parts).replace(tzinfo=local_tz)
        return parsed_dt
    return value


_try_parse_dt = _try_django_parse_dt if _DJANGO else _try_python_parse_dt


def datetime_parser(json_dict: dict) -> dict:
    """
    A function to use as `object_hook` when deserializing JSON that parses
    datetime strings to timezone-aware datetime objects.
    :param json_dict: the original `json_dict` to process
    :return: the modified `json_dict`
    """
    for key, value in json_dict.items():
        if isinstance(value, str):
            json_dict[key] = _try_parse_dt(value)
    return json_dict
